pub mod dns;
pub mod domains;

use dns::{RRSet, RRSetPatch, Record};
use domains::Domain;
use reqwest::{Client, Method, RequestBuilder};
use serde::{de::DeserializeOwned, Deserialize, Serialize};

const BASE_URL: &str = "https://desec.io/api/v1/";

#[derive(Clone)]
pub struct Session {
    client: Client,
    login: Login,
}

#[derive(Deserialize, Clone)]
struct Login {
    token: String,
}
#[derive(Serialize)]
struct Req<'a> {
    email: &'a str,
    password: &'a str,
}

#[derive(Debug)]
pub enum DeError {
    RequestError(reqwest::Error),
    InvalidStatus(u16, Option<String>),
    InvalidCredentials,
    InvalidToken,
}

impl std::fmt::Display for DeError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            DeError::RequestError(err) => write!(f, "request error: {err}"),
            DeError::InvalidStatus(code, content) => {
                write!(f, "Unexpected status code [{code}], body: {:?}", content)
            }
            DeError::InvalidCredentials => write!(f, "invalid credentials"),
            DeError::InvalidToken => write!(f, "invalid or expired token"),
        }
    }
}

impl std::error::Error for DeError {}

impl From<reqwest::Error> for DeError {
    fn from(value: reqwest::Error) -> Self {
        Self::RequestError(value)
    }
}

type Result<T> = std::result::Result<T, DeError>;

impl Session {
    pub async fn login(email: &str, password: &str) -> Result<Self> {
        let client = reqwest::Client::new();
        let response = client
            .post(format!("{BASE_URL}/auth/login/"))
            .header("Content-Type", "application/json")
            .json(&Req { email, password })
            .send()
            .await?;

        match response.status().as_u16() {
            200 => Ok(Self {
                client,
                login: response.json().await?,
            }),
            403 => Err(DeError::InvalidCredentials),
            code => Err(DeError::InvalidStatus(code, response.text().await.ok())),
        }
    }

    pub fn with_token(token: String) -> Self {
        Self {
            client: reqwest::Client::new(),
            login: Login { token },
        }
    }

    async fn authorized_query<D>(&self, method: Method, path: &str) -> Result<D>
    where
        D: DeserializeOwned,
    {
        let response = self
            .client
            .request(method, format!("{BASE_URL}/{path}/"))
            .header("Authorization", format!("Token {}", &self.login.token))
            .send()
            .await?;

        match response.status().as_u16() {
            200 => Ok(response.json().await?),
            403 => Err(DeError::InvalidToken),
            code => Err(DeError::InvalidStatus(code, response.text().await.ok())),
        }
    }

    async fn authorized_exchange<R, D>(&self, method: Method, path: &str, item: &R) -> Result<D>
    where
        R: Serialize,
        D: DeserializeOwned,
    {
        let response = self
            .client
            .request(method, format!("{BASE_URL}/{path}/"))
            .header("Authorization", format!("Token {}", &self.login.token))
            .json(item)
            .send()
            .await?;

        match response.status().as_u16() {
            200 | 201 => Ok(response.json().await?),
            403 => Err(DeError::InvalidToken),
            code => Err(DeError::InvalidStatus(code, response.text().await.ok())),
        }
    }

    async fn authorized_get_query<D, Q>(&self, path: &str, query: Option<Q>) -> Result<D>
    where
        D: DeserializeOwned,
        Q: Serialize,
    {
        let mut request: RequestBuilder = self
            .client
            .get(format!("{BASE_URL}/{path}/"))
            .header("Authorization", format!("Token {}", &self.login.token));
        if let Some(query) = query {
            request = request.query(&query);
        }
        let response = request.send().await?;

        match response.status().as_u16() {
            200 => Ok(response.json().await?),
            403 => Err(DeError::InvalidToken),
            code => Err(DeError::InvalidStatus(code, response.text().await.ok())),
        }
    }

    pub async fn get_domains(&self) -> Result<Vec<Domain>> {
        self.authorized_query(Method::GET, "domains").await
    }

    pub async fn get_domain(&self, domain: &str) -> Result<Domain> {
        self.authorized_query(Method::GET, &format!("domains/{domain}"))
            .await
    }

    pub async fn get_rrsets(
        &self,
        domain: &str,
        type_filter: Option<Vec<Record>>,
    ) -> Result<Vec<RRSet>> {
        self.authorized_get_query::<Vec<RRSet>, Vec<(&str, Record)>>(
            &format!("domains/{domain}/rrsets"),
            type_filter.map(|f| f.into_iter().map(|rec| ("type", rec)).collect()),
        )
        .await
    }

    pub async fn modify_rrset(&self, rrset: RRSet, patch: RRSetPatch) -> Result<RRSet> {
        self.authorized_exchange(Method::PATCH, &rrset_url(&rrset), &patch)
            .await
    }

    pub async fn create_rrset(&self, rrset: RRSet) -> Result<RRSet> {
        let mut rrset = rrset;
        rrset.subname = Some(rrset.subname.unwrap_or(String::new()));

        self.authorized_exchange(
            Method::POST,
            &format!("domains/{}/rrsets", rrset.domain),
            &rrset,
        )
        .await
    }

    pub async fn delete_rrset(&self, rrset: RRSet) -> Result<()> {
        let response = self
            .client
            .delete(format!("{BASE_URL}/{}/", rrset_url(&rrset)))
            .header("Authorization", format!("Token {}", &self.login.token))
            .send()
            .await?;

        match response.status().as_u16() {
            204 => Ok(()),
            403 => Err(DeError::InvalidToken),
            code => Err(DeError::InvalidStatus(code, response.text().await.ok())),
        }
    }
}

fn rrset_url(rrset: &RRSet) -> String {
    format!(
        "domains/{}/rrsets/{}/{}",
        rrset.domain,
        rrset.subname.clone().unwrap_or("...".into()),
        rrset.record
    )
}
