use std::{
    env,
    fs::File,
    io::{prelude::Write, Read},
    net::ToSocketAddrs,
    os::unix::prelude::PermissionsExt,
    time::Duration,
    vec,
};

use desec::dns::{RRSet, RRSetPatch, Record};
use instant_acme::{
    Account, AuthorizationStatus, ChallengeType, Identifier, LetsEncrypt, NewAccount, NewOrder,
    Order, OrderStatus,
};
use log::{debug, error, info};
use rcgen::{Certificate, CertificateParams, DistinguishedName};
use rsdns::{
    clients::{tokio::Client, ClientConfig},
    constants::Class,
    records::data,
};
use serde::{Deserialize, Serialize};
use tokio::io::AsyncWriteExt;
use tokio_rustls::rustls;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub enum CertStore {
    Provision,
    Existing(CertPaths),
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Config {
    pub domain: String,
    pub subdomain: Option<String>,
    pub insecure_port: u16,
    pub tls_port: u16,
    pub cert_store: CertStore,
    pub desec_cfg: DesecConfig,
    #[serde(skip)]
    original_path: &'static str,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct DesecConfig {
    pub username: String,
    #[serde(skip)]
    pub password: Option<String>,
    pub name_servers: Vec<String>,
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct CertPaths {
    pub cert_chain_path: String,
    pub private_key_path: String,
}

impl CertPaths {
    pub fn into_config(&self) -> Result<rustls::ServerConfig, anyhow::Error> {
        let mut buf = std::io::BufReader::new(File::open(&self.cert_chain_path)?);
        let certs = rustls_pemfile::certs(&mut buf)?
            .into_iter()
            .map(tokio_rustls::rustls::Certificate)
            .collect();

        let mut reader = std::io::BufReader::new(File::open(&self.private_key_path)?);
        let mut keys = rustls_pemfile::pkcs8_private_keys(&mut reader)?;
        let key = match keys.len() {
            0 => {
                return Err(anyhow::anyhow!(
                    "No PKCS8-encoded private key found in {}",
                    &self.private_key_path,
                ))
            }
            1 => rustls::PrivateKey(keys.remove(0)),
            _ => {
                return Err(anyhow::anyhow!(
                    "More than one PKCS8-encoded private key found in {}",
                    &self.private_key_path,
                ))
            }
        };

        Ok(rustls::ServerConfig::builder()
            .with_safe_defaults()
            .with_no_client_auth()
            .with_single_cert(certs, key)?)
    }
}

const CONFIG_PATHS: [&str; 3] = [DEFAULT_PATH, "/etc/salut.toml", "/usr/local/etc/salut.toml"];
const CERT_PATHS: [&str; 3] = ["/etc/salut/", "/usr/local/etc/salut/", "./"];
const CERT_FILENAME: &str = "salut.cert";
const CERT_PRIVKEY_FILENAME: &str = "salut.pk";
const DNS_QUERY_WAIT: Duration = Duration::from_millis(250);
const DNS_PASSWORD_VAR: &str = "DESEC_PASSWORD";
pub const DEFAULT_PATH: &str = "salut.toml";

impl Default for Config {
    fn default() -> Self {
        Self {
            domain: String::new(),
            subdomain: Some(String::new()),
            insecure_port: 5222,
            tls_port: 5223,
            cert_store: CertStore::Provision,
            desec_cfg: DesecConfig {
                username: String::new(),
                password: None,
                name_servers: vec!["ns1.desec.io".into(), "ns2.desec.org".into()],
            },
            original_path: DEFAULT_PATH,
        }
    }
}

impl Config {
    pub fn load() -> Result<Self, anyhow::Error> {
        for path in CONFIG_PATHS {
            if let Ok(mut file) = File::open(path) {
                let mut cfg = String::new();
                file.read_to_string(&mut cfg)?;
                let mut cfg: Self = toml::from_str(&cfg)?;
                cfg.original_path = path;
                if let Ok(pass) = env::var(DNS_PASSWORD_VAR) {
                    cfg.desec_cfg.password = Some(pass);
                }
                return Ok(cfg);
            };
        }
        Err(anyhow::anyhow!(
            "could not find salut.toml in {CONFIG_PATHS:?}"
        ))
    }

    pub fn save(&self, path: &str) -> Result<(), anyhow::Error> {
        Ok(write!(
            File::create(path)?,
            "{}",
            toml::to_string_pretty(self)?
        )?)
    }

    pub fn hostname(&self) -> String {
        match &self.subdomain {
            Some(sub) => format!("{sub}.{}", &self.domain),
            None => self.domain.clone(),
        }
    }
}

const ACME_PREFIX: &str = "_acme-challenge";

impl Config {
    // Returns existing certificate or provisions a new one via DNS challenge using DeSEC
    pub async fn certificate(&self) -> Result<CertPaths, anyhow::Error> {
        let desec_cfg = match self.cert_store.clone() {
            CertStore::Provision => self.desec_cfg.clone(),
            CertStore::Existing(existing) => return Ok(existing),
        };

        let account = Account::create(
            &NewAccount {
                contact: &["mailto:emilis@puff.place"],
                terms_of_service_agreed: true,
                only_return_existing: false,
            },
            LetsEncrypt::Production.url(),
            None,
        )
        .await?;

        let identifier = Identifier::Dns(self.hostname());
        let mut order = account
            .new_order(&NewOrder {
                identifiers: &[identifier],
            })
            .await?;

        debug!("order url: {}", order.url());
        let state = order.state();
        info!("cert order state: {:#?}", state);
        assert!(matches!(state.status, OrderStatus::Pending));

        debug!("logging into desec as <{}>", &desec_cfg.username);
        let desec_pass = match desec_cfg.password {
            Some(pass) => pass,
            None => panic!("need desec password for provisioning the TLS certificate, please set {DNS_PASSWORD_VAR}"),
        };
        let dns = desec::Session::login(&desec_cfg.username, &desec_pass).await?;
        debug!("querying existing TXT records");
        let existing_records: Vec<RRSet> = dns
            .get_rrsets(&self.domain, Some(vec![Record::TXT]))
            .await?
            .into_iter()
            .filter(|rec| {
                if let Some(sub) = &rec.subname {
                    sub.starts_with(ACME_PREFIX)
                } else {
                    false
                }
            })
            .collect();
        debug!(
            "got {} existing DNS TXT records that match {ACME_PREFIX}",
            existing_records.len()
        );

        let authorizations = order.authorizations().await.unwrap();
        debug!("got {} authorizations for this order", authorizations.len());
        let mut challenges = Vec::with_capacity(authorizations.len());
        let mut cleanup_records = vec![];

        for authz in &authorizations {
            if let AuthorizationStatus::Valid = authz.status {
                debug!("Valid authorization, skipping: {authz:?}");
                continue;
            }

            // We'll use the DNS challenges for this example, but you could
            // pick something else to use here.
            let challenge = authz
                .challenges
                .iter()
                .find(|c| c.r#type == ChallengeType::Dns01)
                .ok_or_else(|| anyhow::anyhow!("no dns01 challenge found"))?;

            let Identifier::Dns(identifier) = &authz.identifier;
            let dns_challenge = order.key_authorization(challenge).dns_value();
            let subname = format!(
                "{ACME_PREFIX}{}",
                match &self.subdomain {
                    Some(sub) => ".".to_owned() + sub,
                    None => String::new(),
                }
            );
            debug!("challenge for {identifier} ready: {dns_challenge}");

            let record = if let Some(id) = (&existing_records)
                .into_iter()
                .find(|r| *r.subname.as_ref().unwrap() == subname)
            {
                debug!("modifying existing record: {}", id.name);
                dns.modify_rrset(
                    id.clone(),
                    RRSetPatch {
                        name: Some(self.hostname()),
                        subname: Some(subname),
                        records: Some(vec![format!("\"{dns_challenge}\"")]),
                        ..RRSetPatch::default()
                    },
                )
                .await?;
                id.clone()
            } else {
                let record = RRSet {
                    record: Record::TXT,
                    domain: self.domain.clone(),
                    name: self.hostname(),
                    subname: Some(subname),
                    ttl: 3600,
                    records: vec![format!("\"{dns_challenge}\"")],
                    created: String::new(),
                    touched: String::new(),
                };
                debug!("creating new record: {record:?}");
                dns.create_rrset(record).await?
            };

            challenges.push((identifier, &challenge.url));
            cleanup_records.push((record, dns_challenge));
        }
        let challenges_result = self
            .wait_challenges(&cleanup_records, &challenges, &mut order)
            .await;

        for (rec, _) in cleanup_records {
            if let Err(err) = dns.delete_rrset(rec).await {
                error!("failed cleaning up record: {err}")
            }
        }

        challenges_result?;

        let mut names = Vec::with_capacity(challenges.len());
        for (identifier, _) in challenges {
            names.push(identifier.to_owned());
        }

        // If the order is ready, we can provision the certificate.
        // Use the rcgen library to create a Certificate Signing Request.

        let mut params = CertificateParams::new(names.clone());
        params.distinguished_name = DistinguishedName::new();
        let cert = Certificate::from_params(params).unwrap();
        let csr = cert.serialize_request_der()?;

        // Finalize the order and update config
        order.finalize(&csr).await.unwrap();
        let cert_chain_pem = loop {
            match order.certificate().await.unwrap() {
                Some(cert_chain_pem) => break cert_chain_pem,
                None => tokio::time::sleep(Duration::from_secs(1)).await,
            }
        };

        let mut new_cfg = self.clone();
        let paths = save_certs_pam(cert_chain_pem, cert.serialize_private_key_pem()).await?;
        new_cfg.cert_store = CertStore::Existing(paths.clone());
        new_cfg.save(self.original_path)?;

        Ok(paths)
    }

    async fn wait_challenges(
        &self,
        cleanup_records: &Vec<(RRSet, String)>,
        challenges: &Vec<(&String, &String)>,
        order: &mut Order,
    ) -> Result<(), anyhow::Error> {
        for (record, dns_challenge) in cleanup_records {
            for ns in &self.desec_cfg.name_servers {
                info!("waiting for nameserver {ns}");
                loop {
                    let ns = format!("{ns}:53");
                    let mut client = Client::new(ClientConfig::with_nameserver(
                        ns.to_socket_addrs()?
                            .as_slice()
                            .first()
                            .ok_or(anyhow::anyhow!("cannot get address for {ns}"))?
                            .to_owned(),
                    ))
                    .await?;
                    let record = match client
                        .query_rrset::<data::Txt>(&record.name, Class::In)
                        .await
                    {
                        Ok(r) => r,
                        Err(err) => {
                            if let rsdns::Error::NoAnswer = err {
                                tokio::time::sleep(DNS_QUERY_WAIT).await;
                                continue;
                            } else {
                                return Err(err.into());
                            }
                        }
                    };
                    debug!(
                        "{ns} came back with record: {:?}",
                        (&record.rdata)
                            .into_iter()
                            .map(|r| String::from_utf8(r.text.clone()).unwrap_or_default())
                            .collect::<Vec<String>>()
                    );
                    if record
                        .rdata
                        .into_iter()
                        .any(|r| r.text.eq_ignore_ascii_case(dns_challenge.as_bytes()))
                    {
                        break;
                    }
                    tokio::time::sleep(DNS_QUERY_WAIT).await;
                }
            }
        }
        info!("all nameservers are serving (slay), waiting 3 more seconds");
        tokio::time::sleep(Duration::from_secs(3)).await;
        info!("notifying CA");

        for (_, url) in challenges {
            order.set_challenge_ready(url).await.unwrap();
        }

        let mut tries = 1u8;
        let mut delay = Duration::from_millis(250);
        loop {
            tokio::time::sleep(delay).await;
            let state = order.refresh().await?;
            if let OrderStatus::Ready | OrderStatus::Invalid = state.status {
                info!("order state: {:#?}", state);
                break;
            } else {
                info!("waiting on order... state: {:?}", state.status);
            }

            delay *= 2;
            tries += 1;
            match tries < 5 {
                true => info!("[{state:?}({tries})] order is not ready, waiting {delay:?}"),
                false => {
                    info!("[{state:?}({tries})] order is not ready");
                    return Err(anyhow::anyhow!("order is not ready"));
                }
            }
        }

        let state = order.state();
        match &state.status {
            OrderStatus::Ready => Ok(()),
            status => Err(anyhow::anyhow!(
                "unexpected order status: {:?}\nwith state:{state:#?}",
                status
            )),
        }
    }
}
async fn save_certs_pam(
    cert_chain: String,
    private_key: String,
) -> Result<CertPaths, anyhow::Error> {
    for try_path in CERT_PATHS {
        if let Err(_) = std::fs::create_dir_all(try_path) {
            continue;
        }
        let cert_chain_path = format!("{try_path}{CERT_FILENAME}");
        let mut cert_file = match tokio::fs::File::create(&cert_chain_path).await {
            Ok(f) => f,
            Err(_) => continue,
        };
        let private_key_path = format!("{try_path}{CERT_PRIVKEY_FILENAME}");
        let mut pk_file = tokio::fs::File::create(&private_key_path).await?;
        pk_file.metadata().await?.permissions().set_mode(600);
        cert_file.metadata().await?.permissions().set_mode(600);

        cert_file.write_all(cert_chain.as_bytes()).await?;
        pk_file.write_all(private_key.as_bytes()).await?;
        cert_file.flush().await?;
        pk_file.flush().await?;

        return Ok(CertPaths {
            cert_chain_path,
            private_key_path,
        });
    }

    Err(anyhow::anyhow!(
        "could not create/save cert files in {CERT_PATHS:?}"
    ))
}
