use std::{net::SocketAddr, sync::Arc};

use log::{debug, error, info};
use quick_xml::{
    events::{BytesEnd, BytesStart, Event},
    Reader, Writer,
};
use tokio::{
    io::{AsyncWriteExt, BufReader, ReadHalf, WriteHalf},
    net::TcpStream,
};
use tokio_rustls::rustls;

use crate::{
    common::{self, StreamHeader, XMLNamespace},
    error::StreamError,
    feature::Feature,
    negotiator::{self, TcpConnOrTLS},
    tag::{self, Tag},
    tls::stream,
};

pub(crate) enum StartTLSResult {
    Success(stream::TLSStream),
    Failure(StreamStart, StreamError),
    /// TLSFailure is returned when the TLS negotiation failed, and the TCP
    /// stream is to be dropped.
    TLSFailure(StreamError),
}

type Result<T> = std::result::Result<T, StreamError>;

const FEATURES: &'static [Feature] = &[Feature::start_tls(true)];

pub struct StreamStart {
    id: String,
    reader: Reader<BufReader<ReadHalf<TcpStream>>>,
    writer: Writer<WriteHalf<TcpStream>>,
    buffer: Vec<u8>,
    hostname: String,
    tls_config: Arc<rustls::ServerConfig>,
}

impl StreamStart {
    fn new(tcp_stream: TcpStream, hostname: String, tls_config: Arc<rustls::ServerConfig>) -> Self {
        let (read, write) = tokio::io::split(tcp_stream);
        let (reader, writer) = (
            Reader::from_reader(BufReader::new(read)),
            Writer::new(write),
        );

        Self {
            reader,
            writer,
            hostname,
            tls_config,
            id: common::gen_id(),
            buffer: vec![],
        }
    }

    async fn start_stream(self) {
        match self.negotiate_stream().await {
            StartTLSResult::Success(tls_stream) => tls_stream.start_stream().await,
            StartTLSResult::Failure(mut conn, err) => {
                if let Err(err2) = common::error(conn.writer.get_mut(), err).await {
                    error!("error writing error: {err2}");
                    return;
                } else {
                    info!("wrote error {err}")
                }

                if let Err(e) = conn.writer.get_mut().write_all(b"</stream:stream>").await {
                    error!("writing end to stream: {e}")
                }
                if let Err(e) = conn.writer.get_mut().shutdown().await {
                    error!("shutting down stream: {e}")
                }
            }
            StartTLSResult::TLSFailure(err) => {
                error!("TLS negotiation failure: {err}. Dropping connection.")
            }
        }
    }

    async fn negotiate_stream(mut self) -> StartTLSResult {
        let attrs = loop {
            let event = match self.reader.read_event_into_async(&mut self.buffer).await {
                Ok(event) => event,
                Err(err) => return StartTLSResult::Failure(self, err.into()),
            };
            match event {
                Event::Start(start) => {
                    if start.name().as_ref() == tag::STREAM_ELEMENT_NAME {
                        let attrs: StreamHeader = match start.attributes().try_into() {
                            Ok(a) => a,
                            Err(err) => return StartTLSResult::Failure(self, err.into()),
                        };
                        if attrs.namespace != XMLNamespace::JabberClient {
                            return StartTLSResult::Failure(self, StreamError::InvalidNamespace);
                        }
                        break attrs;
                    } else {
                        info!("element: {:?}", start);
                    }
                }
                Event::End(_) => return StartTLSResult::Failure(self, StreamError::BadFormat),
                Event::Eof => return StartTLSResult::Failure(self, StreamError::BadFormat),
                _ => continue,
            }
        };
        info!("starting negotiation with: {attrs:?}");
        if let Err(err) = common::write_stream_header(
            self.writer.get_mut(),
            StreamHeader {
                from: Some(self.hostname.clone()),
                to: attrs.from,
                id: Some(self.id.clone()),
                namespace: XMLNamespace::JabberClient,
            },
        )
        .await
        {
            return StartTLSResult::Failure(self, err);
        };
        if attrs.to.unwrap_or_default() != self.hostname {
            return StartTLSResult::Failure(self, StreamError::HostUnknown);
        }
        debug!("sending features");
        if let Err(err) = self.send_features().await {
            return StartTLSResult::Failure(self, err);
        };
        loop {
            let event = match self.reader.read_event_into_async(&mut self.buffer).await {
                Ok(event) => event,
                Err(err) => return StartTLSResult::Failure(self, err.into()),
            };
            match event {
                Event::Empty(empty) => match empty.name().as_ref() {
                    tag::STARTTLS => {
                        info!("starttls negotiation");
                        let hostname = self.hostname;
                        let tls_config = self.tls_config.clone();
                        return match negotiator::start_tls(
                            self.reader
                                .into_inner()
                                .into_inner()
                                .unsplit(self.writer.into_inner()),
                            empty,
                            self.tls_config.clone(),
                        )
                        .await
                        {
                            TcpConnOrTLS::TLS(conn) => {
                                StartTLSResult::Success(stream::TLSStream::new(conn, hostname))
                            }
                            TcpConnOrTLS::TCP(conn, err) => StartTLSResult::Failure(
                                StreamStart::new(conn, hostname, tls_config),
                                err,
                            ),
                            TcpConnOrTLS::Failed(err) => StartTLSResult::TLSFailure(err),
                        };
                    }
                    _ => return StartTLSResult::Failure(self, StreamError::UnsupportedFeature),
                },
                Event::End(_) => return StartTLSResult::Failure(self, StreamError::BadFormat),
                Event::Eof => return StartTLSResult::Failure(self, StreamError::BadFormat),
                _ => continue,
            }
        }
    }

    async fn send_features(&mut self) -> Result<()> {
        self.writer
            .write_event_async(Event::Start(BytesStart::new(tag::FEATURE)))
            .await?;
        for feature in FEATURES.into_iter() {
            feature.write_tag(self.writer.get_mut()).await?;
        }
        self.writer
            .write_event_async(Event::End(BytesEnd::new(tag::FEATURE)))
            .await?;
        Ok(())
    }
}

pub fn spawn(
    hostname: String,
    (stream, _): (TcpStream, SocketAddr),
    tls_config: Arc<rustls::ServerConfig>,
) {
    tokio::spawn(async move {
        StreamStart::new(stream, hostname, tls_config)
            .start_stream()
            .await;
    });
}
